/*
 * Real time packet sending client
 *
 * Large portions taken from cyclictest
 *
 */

#define _GNU_SOURCE
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <limits.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "common.h"
#include "send_packet.h"

// Structs

typedef struct thread_param {
  uint64_t interval;
  unsigned int max_cycles;
  int priority;
  int etf_offset;
  int64_t send_ts_delay;
  int affinity_cpu;
  uint64_t start_ts;

} thread_param_t;

typedef struct main_param {
  int refresh_rate;
  int verbose;

  int transition_time;
  uint64_t target_interval;

} main_param_t;

// Static functions

static void process_options(int argc, char *argv[]);
static void do_tsn_task(char *data, uint64_t next_txtime);
static void sighand(int sig_num);

// Static variables

static uint64_t nb_cycles;
static main_param_t main_params;
static thread_param_t thread_params;

static egress_param_t egress_params;
static egress_param_t egress_params2;
static egress_stat_t egress_stats = {.packets_sent = 0,
  .min_interval = INT_MAX};
static egress_stat_t egress_stats2 = {.packets_sent = 0,
  .min_interval = INT_MAX};

static egress_info_t egress_info;
static egress_info_t egress_info2;

static int enable_etf;
static int two_servers = 0;
static struct timespec measures_start;
static struct timespec measures_end;
static char send_data[MAX_BUFFER_SIZE];

static int reverse_motors = 0;

static void help(char *argv[]) {
  printf(
      "Usage: %s [-a CPU -p PRIO -i USEC -r USEC -l N] [-d BUF_LEN | -c "
      "DELAY -s NS] [-b -e OFFSET -q PK_PRIO -gtvT] IF IP IF2 IP2\n\n"
      "       -h                 Show help\n"
      "       -a CPU             Pin the real time thread to CPU\n"
      "       -p PRIO            RT thread priority\n"
      "       -c USEC            Send timestamp with the specified delay\n"
      "       -s NS              Common start time reference\n"
      "       -e USEC            txtime userspace offset (for ETF qdisc)\n"
      "       -q PK_PRIO         Packet priority \n"
      "       -t T               Transition time\n"
      "\n",
    argv[0]);
}

static double i_to_rps(int i) {
  return ((double)USEC_PER_SEC) / (MOTOR_STEPS * i);
}
static int rps_to_i(double rps) {
  return USEC_PER_SEC / (MOTOR_STEPS * rps);
}

static void *print_thread(void *p) {
  (void)p;

  for (;;) {
    if (thread_params.max_cycles &&
        nb_cycles >= ((unsigned int)thread_params.max_cycles))
      break;

    printf("RPS: %5F RPS Target: %5F s/100 RPS %5d\n", i_to_rps(thread_params.interval / 1000),
                                                       i_to_rps(main_params.target_interval),
                                                       main_params.transition_time);
    printf("\033[%dA", 1);

    usleep(100000);
  }

  pthread_exit(NULL);
}

/*
 * Real-time thread:
 *   - Sends packets at a regular intervall
 */
static void *packet_sending_thread(void *p) {
  (void)p;
  struct timespec next, current, previous;
  uint64_t next_txtime = 0, tx_data = 0;
  int ret;
  cpu_set_t mask;

  uint64_t next_increment = 0;
  uint64_t end_t = 0;
  uint64_t cur_t = 0;
  uint64_t i_t = 0;
  uint64_t i_s = thread_params.interval / 1000;
  uint64_t i_c = thread_params.interval / 1000;

  // Set thread CPU affinity
  if (thread_params.affinity_cpu != -1) {
    CPU_ZERO(&mask);
    CPU_SET(thread_params.affinity_cpu, &mask);
    if (sched_setaffinity(0, sizeof(mask), &mask))
      error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n",
          thread_params.affinity_cpu);
  }

  clock_gettime(CLOCK_MONOTONIC, &measures_start);
  clock_gettime(CLOCK_REALTIME, &next);

  if (thread_params.start_ts) {
    if (thread_params.start_ts < ts_to_uint(next)) {
      fprintf(stderr, "start timestamp is in the past, aborting...\n");
      fflush(stdout);
      exit(EXIT_FAILURE);
    }
    if (thread_params.start_ts > (ts_to_uint(next) + UINT64_C(3600000000000))) {
      fprintf(stderr, "start timestamp is too high, aborting...\n");
      fflush(stdout);
      exit(EXIT_FAILURE);
    }
    next = uint_to_ts(thread_params.start_ts);
  }
  add_ns(&next, thread_params.interval);

  // Set txtime and TX data timestamp if required
  if (enable_etf) {
    next_txtime = next.tv_sec * NSEC_PER_SEC + next.tv_nsec;
    next_txtime += thread_params.etf_offset;
  }

  tx_data = next.tv_sec * NSEC_PER_SEC + next.tv_nsec;
  tx_data += thread_params.send_ts_delay;

  printf("Will start at %" PRIu64 "\n", ts_to_uint(next));

  // Packet sending loop
  for (nb_cycles = 0;; nb_cycles++) {
    if (thread_params.max_cycles &&
        nb_cycles >= ((unsigned int)thread_params.max_cycles))
      break;

    clock_gettime(CLOCK_REALTIME, &current);

    // Sanity check
    if(current.tv_sec > next.tv_sec)
      goto invalid_ts;
    if(current.tv_sec == next.tv_sec && current.tv_nsec > next.tv_nsec)
      goto invalid_ts;

    ret = clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &next, NULL);
    if (ret) {
      fprintf(stderr, "clock_nanosleep returned error: %d, aborting...\n", ret);
      exit(EXIT_FAILURE);
    }

    // TX data
    encode(tx_data, send_data);

    // Get timestamp before TSN task for stats
    clock_gettime(CLOCK_REALTIME, &current);
    do_tsn_task(send_data, next_txtime);
    add_ns(&next, thread_params.interval);

    // Update txtime and TX data timestamp if required
    if (enable_etf) next_txtime += thread_params.interval;
    tx_data = next.tv_sec * NSEC_PER_SEC + next.tv_nsec;
    tx_data += thread_params.send_ts_delay;

    if(reverse_motors) {
      reverse_motors = 0;
      tx_data = 17;
    }

    // Update statistics
    if (nb_cycles) {
      int interval_us = calcdiff_ns(current, previous) / 1000;

      egress_stats.min_interval = _min_(interval_us, egress_stats.min_interval);
      egress_stats.max_interval = _max_(interval_us, egress_stats.max_interval);
      egress_stats.avg_interval =
        (egress_stats.avg_interval * nb_cycles + interval_us) /
        (nb_cycles + 1);
    }

    if(thread_params.interval != (main_params.target_interval * 1000)) {

      i_c = thread_params.interval / 1000;

      // If target interval changed
      if(i_t != main_params.target_interval) {

        i_t = main_params.target_interval;
        i_s = i_c;
        cur_t = 0;
        next_increment = 0;

        if(i_t < i_s)
          end_t = (main_params.transition_time * USEC_PER_SEC * USEC_PER_SEC * (i_c - i_t)) / (100 * MOTOR_STEPS * i_t * i_c);
        else
          end_t = (main_params.transition_time * USEC_PER_SEC * USEC_PER_SEC * (i_t - i_c)) / (100 * MOTOR_STEPS * i_t * i_c);

      }

      if(next_increment) {
        if(cur_t > next_increment) {
          i_c += i_t < i_s ? -1 : 1;
          next_increment = 0;
        }
      } else {
        uint64_t next_i = i_t < i_s ? i_c - 1 : i_c + 1;
        // Compute time at which we will need to increment / decrement the interval

        if(i_t < i_s)
          next_increment = (end_t * i_t * (i_s - next_i)) / ((i_s - i_t) * next_i);
        else
          next_increment = (end_t * i_t * (next_i - i_s)) / ((i_t - i_s) * next_i);

        // If next increment time is before next interval
        if(next_increment < cur_t + i_c) {
          if(i_t < i_s) {
            i_c = (i_t * i_s * end_t) / ((i_s - i_t) * (cur_t + i_c) + end_t * i_t);
            i_c = i_c < i_t ? i_t : i_c;
          }
          else {
            i_c = (i_t * i_s * end_t) / (end_t * i_t - (i_t - i_s) * (cur_t + i_c));
            i_c = i_c > i_t ? i_t : i_c;
          }
          next_increment = 0;
        }
      }

      cur_t += i_c;
      if(i_c < 50) {
        fprintf(stderr, "Interval too small, exiting..\n");
        exit(EXIT_FAILURE);
      }
      if(i_c > USEC_PER_SEC) {
        fprintf(stderr, "Interval too big, exiting..\n");
        exit(EXIT_FAILURE);
      }
      thread_params.interval = i_c * 1000;
    }

    previous = current;
  }

  pthread_exit(NULL);

invalid_ts:
  fprintf(stderr, "Invalid timestamp         : %" PRIu64 "\n", ts_to_uint(next));
  fprintf(stderr, "Current time is           : %" PRIu64 "\n", ts_to_uint(current));
  fprintf(stderr, "Difference with timestamp : %" PRIi64 "\n\n\n", ((int64_t)ts_to_uint(next)) - ((int64_t)ts_to_uint(current)));
  exit(EXIT_FAILURE);
}

static void motor_input(void) {

  char user_input[255];
  int v;
  scanf("%s", user_input);
  switch(user_input[0]) {
    case 'a':
      v = atoi(user_input + 1);
      if(v)
        main_params.transition_time = v;
      break;
    case 'r':
      reverse_motors = 1;
      break;
    default:
      v = rps_to_i(strtod(user_input, NULL));
      if(v)
        main_params.target_interval = v;
      break;
  }
}

/*
 * Main thread:
 *  - Has non-real time priority
 *  - Handles the IO and creates the real time thread
 */
int main(int argc, char *argv[]) {
  pthread_t thread, print_pthread;
  struct sched_param param;
  pthread_attr_t attr;

  // Default configuration values
  thread_params.interval = 100000 * 1000;
  thread_params.priority = 98;
  thread_params.send_ts_delay = 0;
  thread_params.affinity_cpu = -1;
  thread_params.start_ts = 0;
  main_params.refresh_rate = 50000;
  main_params.verbose = 0;
  egress_params.packet_priority = 3;
  egress_params.tx_buffer_len = 1024;
  egress_params2.packet_priority = 3;
  egress_params2.tx_buffer_len = 1024;
  enable_etf = 0;

  // Lock all current and future pages from preventing of being paged to
  // swap
  if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
    fprintf(stderr, "mlockall failed");
    exit(EXIT_FAILURE);
  }

  // Process bash options
  process_options(argc, argv);

  main_params.target_interval = thread_params.interval / 1000;

  set_latency_target();

  egress_params.use_etf = enable_etf;
  egress_params2.use_etf = enable_etf;

  egress_info.params = &egress_params;
  egress_info.stats = &egress_stats;
  egress_info2.params = &egress_params2;
  egress_info2.stats = &egress_stats2;

  // Catch breaks with sighand
  init_signals(sighand);

  // Initialize the UDP packet sending socket
  init_udp_send(&egress_info);
  if(two_servers)
    init_udp_send(&egress_info2);

  // Initialize pthread attributes (default values)
  if (pthread_attr_init(&attr)) {
    fprintf(stderr, "init pthread attributes failed\n");
    exit(EXIT_FAILURE);
  }
  // Set a specific stack size
  if (pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN)) {
    fprintf(stderr, "pthread setstacksize failed\n");
    exit(EXIT_FAILURE);
  }
  // Set scheduler policy and priority of pthread
  if (pthread_attr_setschedpolicy(&attr, SCHED_FIFO)) {
    fprintf(stderr, "pthread setschedpolicy failed\n");
    exit(EXIT_FAILURE);
  }
  param.sched_priority = thread_params.priority;
  if (pthread_attr_setschedparam(&attr, &param)) {
    fprintf(stderr, "pthread setschedparam failed\n");
    exit(EXIT_FAILURE);
  }
  /* Use scheduling parameters of attr */
  if (pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED)) {
    fprintf(stderr, "pthread setinheritsched failed\n");
    exit(EXIT_FAILURE);
  }

  // Create the real time thread
  if (pthread_create(&thread, &attr, packet_sending_thread, NULL))
    error(EXIT_FAILURE, errno, "Couldn't create packet sending thread");

  // Create the print thread
  if (pthread_create(&print_pthread, NULL, print_thread, NULL))
    error(EXIT_FAILURE, errno, "Couldn't create print thread");

  // Verbose loop
  for (;;) {
    usleep(main_params.refresh_rate);

    motor_input();

    if (thread_params.max_cycles)
      if (thread_params.max_cycles == nb_cycles) break;
  }

  exit(EXIT_SUCCESS);
}

/* Critical TSN task
*/
static void do_tsn_task(char *data, uint64_t next_txtime) {
  // One way packet sending
  send_udp_packet(data, next_txtime, &egress_info);
  if(two_servers)
    send_udp_packet(data, next_txtime, &egress_info2);
}

static void sighand(int sig_num) {
  (void)sig_num;
  exit(EXIT_SUCCESS);
}

/* Process bash options
*/
static void process_options(int argc, char *argv[]) {
  for (;;) {
    int c = getopt(argc, argv, "a:c:e:hp:q:s:t:");

    if (c == -1) break;

    switch (c) {
      case 'a':
        thread_params.affinity_cpu = atoi(optarg);
        break;
      case 'c':
        thread_params.send_ts_delay = strtoll(optarg, NULL, 10) * 1000;
        break;
      case 'e':
        enable_etf = 1;
        thread_params.etf_offset = atoi(optarg) * 1000;
        break;
      case 'h':
        help(argv);
        exit(EXIT_SUCCESS);
        break;
      case 'p':
        thread_params.priority = atoi(optarg);
        break;
      case 'q':
        egress_params.packet_priority = atoi(optarg);
        break;
      case 's':
        thread_params.start_ts = strtoull(optarg, NULL, 10);
        break;
      case 't':
        main_params.transition_time = atoi(optarg);
        break;
    }
  }

  if (argc != optind + 2 && argc != optind + 4) {
    if (argc < optind + 2)
      fprintf(stderr, "You need to specifiy an interface and IP address\n");
    else
      fprintf(stderr, "Wrong number of arguments");
    help(argv);
    exit(EXIT_FAILURE);
  }

  strcpy(egress_params.network_if, argv[optind]);
  strcpy(egress_params.server_ip, argv[optind + 1]);

  if (argc == optind + 4) {
    two_servers = 1;
    strcpy(egress_params2.network_if, argv[optind + 2]);
    strcpy(egress_params2.server_ip, argv[optind + 3]);
  }
}
