/*
 * Real time packet receiving server
 *
 * Large portions taken from cyclictest
 *
 */

#define _GNU_SOURCE
#include <arpa/inet.h>
#include <errno.h>
#include <error.h>
#include <inttypes.h>
#include <limits.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <unistd.h>

#include "common.h"
#include "gpio.h"
#include "recv_packet.h"

#define ERROR_MARGIN_NS 400000

// Structs

enum TSNTask { RECV_PACKET_TASK, RTT_TASK, XDP_TASK };

typedef struct thread_param {
  int priority;
  int affinity_cpu;
  uint64_t latency_threshold;

  uint64_t start_ts;

} thread_param_t;

typedef struct main_params {
  int refresh_rate;
  int verbose;
} main_param_t;

static void process_options(int argc, char *argv[]);
static void sighand(int sig_num);

// Static variables

static main_param_t main_params;
static thread_param_t thread_params;
static ingress_param_t ingress_params;
static ingress_stat_t ingress_stats = {.min_interval = INT_MAX};
static enum TSNTask tsn_task;
static struct timespec measures_start;
static struct timespec measures_end;
static struct timespec emit_signal_next;
static pthread_mutex_t emit_signal_mutex;
static pthread_cond_t emit_signal_ts_received;

static int interval_us;

static int gpio86_fd;
static int gpio84_fd;
static int gpio86_state;
static int gpio84_state;
static int received_reverse_motor = 0;

static void help(char *argv[]) {
  printf(
      "Usage: %s [-f IF -a CPU -p PRIO -r USEC] [-v] [-x]\n\n"
      "       -h                 Show help\n"
      "       -f IF              Set the network interface to be used\n"
      "       -a CPU             Pin the real time thread to CPU\n"
      "       -p PRIO            RT thread priority\n"
      "       -r USEC            non-RT main thread refresh interval\n"
      "       -s NS              Common start time reference\n"
      "       -x                 Use AF_XDP sockets\n"
      "       -v                 Verbose\n"
      "\n",
      argv[0]);
}

static void *emit_signal_thread(void *p) {
  (void)p;
  cpu_set_t mask;
  struct timespec current;
  struct timespec previous_emit, previous_ts;
  int ret;
  int reverse_motor = 0;

  // Set thread CPU affinity
  if (thread_params.affinity_cpu) {
    CPU_ZERO(&mask);
    CPU_SET(thread_params.affinity_cpu, &mask);
    if (sched_setaffinity(0, sizeof(mask), &mask))
      error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n",
            thread_params.affinity_cpu);
  }

  pthread_mutex_lock(&emit_signal_mutex);

  for (int i = 0;;i++) {
    pthread_cond_wait(&emit_signal_ts_received, &emit_signal_mutex);

    if(received_reverse_motor) {
      received_reverse_motor = 0;
      reverse_motor = 1;
      continue;
    }

    ret = clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &emit_signal_next, NULL);
    if (ret) {
      fprintf(stderr, "clock_nanosleep returned error: %d, aborting...\n", ret);
      exit(EXIT_FAILURE);
    }

    if(reverse_motor) {
      gpio84_state = toggle_gpio(gpio84_fd, gpio84_state);
      reverse_motor = 0;
    }
    else {
      gpio86_state = toggle_gpio(gpio86_fd, gpio86_state);
      usleep(3);
      gpio86_state = toggle_gpio(gpio86_fd, gpio86_state);
    }

    clock_gettime(CLOCK_REALTIME, &current);
    
    previous_emit = current;
    previous_ts = emit_signal_next;
  }

  pthread_mutex_unlock(&emit_signal_mutex);
  pthread_exit(NULL);
}

/* Real-time thread:
 *   - Measures intervals between packet receptions
 */
static void *tsn_thread(void *p) {
  (void)p;
  struct timespec current, previous, next;
  cpu_set_t mask;
  
  // Set thread CPU affinity
  if (thread_params.affinity_cpu) {
    CPU_ZERO(&mask);
    CPU_SET(thread_params.affinity_cpu, &mask);
    if (sched_setaffinity(0, sizeof(mask), &mask))
      error(EXIT_FAILURE, errno, "Could not set CPU affinity to CPU %d\n",
            thread_params.affinity_cpu);
  }

  clock_gettime(CLOCK_MONOTONIC, &measures_start);

  if (tsn_task == XDP_TASK) setup_poll_fd();

  if (thread_params.start_ts) {
    clock_gettime(CLOCK_REALTIME, &next);
    if (thread_params.start_ts < ts_to_uint(next)) {
      fprintf(stderr, "start timestamp is in the past, aborting...\n");
      exit(EXIT_FAILURE);
    }
    if (thread_params.start_ts > (ts_to_uint(next) + UINT64_C(3600000000000))) {
      fprintf(stderr, "start timestamp is too high, aborting...\n");
      exit(EXIT_FAILURE);
    }
    next = uint_to_ts(thread_params.start_ts);
  }

  // Packet receiving loop
  for (ingress_stats.packets_received = 0;; ingress_stats.packets_received++) {

    // Receive UDP or XDP packet
    if (tsn_task == RECV_PACKET_TASK)
      recv_udp_packet();
    else
      recv_xdp_packet();

    // Get time for statistics
    clock_gettime(CLOCK_REALTIME, &current);
    recv_xdp_cleanup();
    interval_us = calcdiff_ns(current, previous) / 1000;

    uint64_t emit_signal_t;

    if(tsn_task == XDP_TASK)
      emit_signal_t = decode(ingress_stats.xdp_data);
    else
      emit_signal_t = decode(ingress_stats.data);

    if(emit_signal_t < UINT64_C(1576800000000000000) && emit_signal_t != 17)
      continue;

    pthread_mutex_lock(&emit_signal_mutex);
    if(emit_signal_t == 17)
      received_reverse_motor = 1;
    else
      emit_signal_next = uint_to_ts(emit_signal_t);
    pthread_cond_signal(&emit_signal_ts_received);
    pthread_mutex_unlock(&emit_signal_mutex);

    // Update stats
    if (ingress_stats.packets_received) {

      ingress_stats.min_interval =
          _min_(interval_us, ingress_stats.min_interval);
      ingress_stats.max_interval =
          _max_(interval_us, ingress_stats.max_interval);
      ingress_stats.avg_interval =
          (ingress_stats.avg_interval * ingress_stats.packets_received +
           interval_us) /
          (ingress_stats.packets_received + 1);
    }

    previous = current;
  }

  return NULL;
}

static void create_thread(void *(*thread_function)(void *)) {
  pthread_t thread;
  pthread_attr_t attr;
  struct sched_param param;

  // Initialize pthread attributes (default values)
  if (pthread_attr_init(&attr)) {
    fprintf(stderr, "init pthread attributes failed\n");
    exit(EXIT_FAILURE);
  }
  // Set a specific stack size
  if (pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN)) {
    fprintf(stderr, "pthread setstacksize failed\n");
    exit(EXIT_FAILURE);
  }
  // Set scheduler policy and priority of pthread
  if (pthread_attr_setschedpolicy(&attr, SCHED_FIFO)) {
    fprintf(stderr, "pthread setschedpolicy failed\n");
    exit(EXIT_FAILURE);
  }
  param.sched_priority = thread_params.priority;
  if (pthread_attr_setschedparam(&attr, &param)) {
    fprintf(stderr, "pthread setschedparam failed\n");
    exit(EXIT_FAILURE);
  }
  // Use scheduling parameters of attr
  if (pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED)) {
    fprintf(stderr, "pthread setinheritsched failed\n");
    exit(EXIT_FAILURE);
  }

  // Create the real time thread
  if (pthread_create(&thread, &attr, thread_function, NULL))
    error(EXIT_FAILURE, errno, "Couldn't create packet receiving thread");
}

/* Main thread:
 *   - Has non-real time priority
 *   - Handles the IO and creates real time threads
 */
int main(int argc, char *argv[]) {
  // Initial values
  ingress_stats.min_interval = INT_MAX;
  ingress_stats.avg_interval = 0;
  ingress_stats.max_interval = 0;
  ingress_stats.packets_received = 0;

  // Default configuration values
  thread_params.priority = 98;
  thread_params.affinity_cpu = 0;
  thread_params.start_ts = 0;
  main_params.refresh_rate = 50000;
  main_params.verbose = 0;
  tsn_task = RECV_PACKET_TASK;
  ingress_params.tx_buffer_len = 1024;

  // Process bash options
  process_options(argc, argv);

  set_latency_target();

  // Catch breaks with sighand to exit XDP
  init_signals(sighand);

  // Initialize the XDP or UDP packet receiving socket
  if (tsn_task == XDP_TASK)
    init_xdp_recv(&ingress_params, &ingress_stats);
  else
    init_udp_recv(&ingress_params, &ingress_stats);

  pthread_mutex_init(&emit_signal_mutex, NULL);
  pthread_cond_init(&emit_signal_ts_received, NULL);

  gpio86_state = enable_gpio(&gpio86_fd, 86);
  gpio84_state = enable_gpio(&gpio84_fd, 84);

  create_thread(tsn_thread);
  create_thread(emit_signal_thread);

  // Verbose loop
  for (;;) {
    usleep(main_params.refresh_rate);

    // Stats printing
    if (main_params.verbose && ingress_stats.packets_received > 1) {
      int jitter = ingress_stats.max_interval - ingress_stats.min_interval;

      printf("%9" PRIu64 ": J: %5d, I (10us): %3d %3d %3d [%3d]",
             ingress_stats.packets_received, jitter,
             ingress_stats.min_interval / 10, ingress_stats.avg_interval / 10,
             ingress_stats.max_interval / 10, (int)ingress_stats.high_jitter);

      printf("\n");
      printf("\033[%dA", 1);
    }
  }

  pthread_mutex_destroy(&emit_signal_mutex);
  pthread_cond_destroy(&emit_signal_ts_received);
  close_xdp_socket();
}

static void sighand(int sig_num) {
  (void)sig_num;

  if (tsn_task == XDP_TASK) {
    close_xdp_socket();
  }

  if (ingress_stats.high_jitter)
    fprintf(stderr, "%d packets were lost\n", ((int)ingress_stats.high_jitter));

  exit(EXIT_SUCCESS);
}

/* Process bash options
 */
static void process_options(int argc, char *argv[]) {
  int network_if_specified = 0;

  for (;;) {
    int c = getopt(argc, argv, "a:s:f:hp:r:vx");

    if (c == -1) break;

    switch (c) {
      case 'a':
        thread_params.affinity_cpu = atoi(optarg);
        break;
      case 's':
        thread_params.start_ts = strtoull(optarg, NULL, 10);
        break;
      case 'f':
        network_if_specified = 1;
        strcpy(ingress_params.network_if, optarg);
        break;
      case 'h':
        help(argv);
        exit(EXIT_SUCCESS);
        break;
      case 'p':
        thread_params.priority = atoi(optarg);
        break;
      case 'r':
        main_params.refresh_rate = atoi(optarg);
        break;
      case 'v':
        main_params.verbose = 1;
        break;
      case 'x':
        tsn_task = XDP_TASK;
        break;
    }
  }

  if (!network_if_specified) {
    fprintf(stderr, "You need to specifiy an network interface\n");
    help(argv);
    exit(EXIT_FAILURE);
  }

  if (argc != optind) {
    fprintf(stderr, "Too many arguments\n");
    help(argv);
    exit(EXIT_FAILURE);
  }
}
